@@ -15,9 +15,12 @@
 package panos
 
 import (
+	"encoding/base64"
 	"fmt"
+	"strconv"
 
 	"github.com/GoogleCloudPlatform/terraformer/terraformutils"
+	"github.com/PaloAltoNetworks/pango"
 	"github.com/PaloAltoNetworks/pango/netw/interface/eth"
 	"github.com/PaloAltoNetworks/pango/netw/interface/subinterface/layer2"
 	"github.com/PaloAltoNetworks/pango/netw/interface/subinterface/layer3"
@@ -44,13 +47,13 @@ func (g *FirewallNetworkingGenerator) createResourcesFromList(o getGeneric, idPr
 	default:
 		err = fmt.Errorf("not supported")
 	}
-	if err != nil {
+	if err != nil || len(l) == 0 {
 		return []terraformutils.Resource{}
 	}
 
 	for _, r := range l {
 		if checkIfIsVsys {
-			rv, err := g.client.IsImported(checkType, "", "", g.vsys, r)
+			rv, err := g.client.(*pango.Firewall).IsImported(checkType, "", "", g.vsys, r)
 			if err != nil || !rv {
 				continue
 			}
@@ -76,13 +79,13 @@ func (g *FirewallNetworkingGenerator) createResourcesFromList(o getGeneric, idPr
 }
 
 func (g *FirewallNetworkingGenerator) createAggregateInterfaceResources() (resources []terraformutils.Resource) {
-	l, err := g.client.Network.AggregateInterface.GetList()
+	l, err := g.client.(*pango.Firewall).Network.AggregateInterface.GetList()
 	if err != nil {
 		return []terraformutils.Resource{}
 	}
 
 	for _, aggregateInterface := range l {
-		rv, err := g.client.IsImported(util.InterfaceImport, "", "", g.vsys, aggregateInterface)
+		rv, err := g.client.(*pango.Firewall).IsImported(util.InterfaceImport, "", "", g.vsys, aggregateInterface)
 		if err != nil || !rv {
 			continue
 		}
@@ -96,7 +99,7 @@ func (g *FirewallNetworkingGenerator) createAggregateInterfaceResources() (resou
 			[]string{},
 		))
 
-		e, err := g.client.Network.AggregateInterface.Get(aggregateInterface)
+		e, err := g.client.(*pango.Firewall).Network.AggregateInterface.Get(aggregateInterface)
 		if err != nil {
 			continue
 		}
@@ -115,7 +118,7 @@ func (g *FirewallNetworkingGenerator) createAggregateInterfaceResources() (resou
 
 func (g *FirewallNetworkingGenerator) createBFDProfileResources() []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.BfdProfile, []string{}},
+		getGeneric{g.client.(*pango.Firewall).Network.BfdProfile, []string{}},
 		"", false, "panos_bfd_profile", false, "",
 	)
 }
@@ -131,7 +134,7 @@ func (g *FirewallNetworkingGenerator) createBGPResource(virtualRouter string) te
 }
 
 func (g *FirewallNetworkingGenerator) createBGPAggregateResources(virtualRouter string) (resources []terraformutils.Resource) {
-	l, err := g.client.Network.BgpAggregate.GetList(virtualRouter)
+	l, err := g.client.(*pango.Firewall).Network.BgpAggregate.GetList(virtualRouter)
 	if err != nil {
 		return []terraformutils.Resource{}
 	}
@@ -155,14 +158,14 @@ func (g *FirewallNetworkingGenerator) createBGPAggregateResources(virtualRouter
 
 func (g *FirewallNetworkingGenerator) createBGPAggregateAdvertiseFilterResources(virtualRouter, bgpAggregate string) []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.BgpAggAdvertiseFilter, []string{virtualRouter, bgpAggregate}},
+		getGeneric{g.client.(*pango.Firewall).Network.BgpAggAdvertiseFilter, []string{virtualRouter, bgpAggregate}},
 		virtualRouter+":"+bgpAggregate+":", true, "panos_bgp_aggregate_advertise_filter", false, "",
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createBGPAggregateSuppressFilterResources(virtualRouter, bgpAggregate string) []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.BgpAggSuppressFilter, []string{virtualRouter, bgpAggregate}},
+		getGeneric{g.client.(*pango.Firewall).Network.BgpAggSuppressFilter, []string{virtualRouter, bgpAggregate}},
 		virtualRouter+":"+bgpAggregate+":", true, "panos_bgp_aggregate_suppress_filter", false, "",
 	)
 }
@@ -170,13 +173,13 @@ func (g *FirewallNetworkingGenerator) createBGPAggregateSuppressFilterResources(
 // The secret argument will contain "(incorrect)", not the real value
 func (g *FirewallNetworkingGenerator) createBGPAuthProfileResources(virtualRouter string) []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.BgpAuthProfile, []string{virtualRouter}},
+		getGeneric{g.client.(*pango.Firewall).Network.BgpAuthProfile, []string{virtualRouter}},
 		virtualRouter+":", true, "panos_bgp_auth_profile", false, "",
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createBGPConditionalAdvertisementResources(virtualRouter string) (resources []terraformutils.Resource) {
-	l, err := g.client.Network.BgpConditionalAdv.GetList(virtualRouter)
+	l, err := g.client.(*pango.Firewall).Network.BgpConditionalAdv.GetList(virtualRouter)
 	if err != nil {
 		return []terraformutils.Resource{}
 	}
@@ -200,41 +203,70 @@ func (g *FirewallNetworkingGenerator) createBGPConditionalAdvertisementResources
 
 func (g *FirewallNetworkingGenerator) createBGPConditionalAdvertisementAdvertiseFilterResources(virtualRouter, bgpConditionalAdv string) []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.BgpConAdvAdvertiseFilter, []string{virtualRouter, bgpConditionalAdv}},
+		getGeneric{g.client.(*pango.Firewall).Network.BgpConAdvAdvertiseFilter, []string{virtualRouter, bgpConditionalAdv}},
 		virtualRouter+":"+bgpConditionalAdv+":", true, "panos_bgp_conditional_adv_advertise_filter", false, "",
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createBGPConditionalAdvertisementNonExistFilterResources(virtualRouter, bgpConditionalAdv string) []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.BgpConAdvNonExistFilter, []string{virtualRouter, bgpConditionalAdv}},
+		getGeneric{g.client.(*pango.Firewall).Network.BgpConAdvNonExistFilter, []string{virtualRouter, bgpConditionalAdv}},
 		virtualRouter+":"+bgpConditionalAdv+":", true, "panos_bgp_conditional_adv_non_exist_filter", false, "",
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createBGPDampeningProfileResources(virtualRouter string) []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.BgpDampeningProfile, []string{virtualRouter}},
+		getGeneric{g.client.(*pango.Firewall).Network.BgpDampeningProfile, []string{virtualRouter}},
 		virtualRouter+":", true, "panos_bgp_dampening_profile", false, "",
 	)
 }
 
+func (g *FirewallNetworkingGenerator) createBGPRuleGroupResourcesFromList(o getGeneric, terraformResourceName string) (resources []terraformutils.Resource) {
+	l, err := o.i.(getListWithOneArg).GetList(o.params[0])
+	if err != nil || len(l) == 0 {
+		return []terraformutils.Resource{}
+	}
+
+	var positionReference string
+	id := o.params[0] + ":" + strconv.Itoa(util.MoveTop) + "::"
+
+	for k, r := range l {
+		if k > 0 {
+			id = o.params[0] + ":" + strconv.Itoa(util.MoveAfter) + ":" + positionReference + ":"
+		}
+
+		id += base64.StdEncoding.EncodeToString([]byte(r))
+		positionReference = r
+
+		resources = append(resources, terraformutils.NewSimpleResource(
+			id,
+			normalizeResourceName(r),
+			terraformResourceName,
+			"panos",
+			[]string{},
+		))
+	}
+
+	return resources
+}
+
 func (g *FirewallNetworkingGenerator) createBGPExportRuleGroupResources(virtualRouter string) []terraformutils.Resource {
-	return g.createResourcesFromList(
-		getGeneric{g.client.Network.BgpExport, []string{virtualRouter}},
-		virtualRouter+":", true, "panos_bgp_export_rule_group", false, "",
+	return g.createBGPRuleGroupResourcesFromList(
+		getGeneric{g.client.(*pango.Firewall).Network.BgpExport, []string{virtualRouter}},
+		"panos_bgp_export_rule_group",
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createBGPImportRuleGroupResources(virtualRouter string) []terraformutils.Resource {
-	return g.createResourcesFromList(
-		getGeneric{g.client.Network.BgpImport, []string{virtualRouter}},
-		virtualRouter+":", true, "panos_bgp_import_rule_group", false, "",
+	return g.createBGPRuleGroupResourcesFromList(
+		getGeneric{g.client.(*pango.Firewall).Network.BgpImport, []string{virtualRouter}},
+		"panos_bgp_import_rule_group",
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createBGPPeerGroupResources(virtualRouter string) (resources []terraformutils.Resource) {
-	l, err := g.client.Network.BgpPeerGroup.GetList(virtualRouter)
+	l, err := g.client.(*pango.Firewall).Network.BgpPeerGroup.GetList(virtualRouter)
 	if err != nil {
 		return []terraformutils.Resource{}
 	}
@@ -257,26 +289,26 @@ func (g *FirewallNetworkingGenerator) createBGPPeerGroupResources(virtualRouter
 
 func (g *FirewallNetworkingGenerator) createBGPPeerResources(virtualRouter, bgpPeerGroup string) []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.BgpPeer, []string{virtualRouter, bgpPeerGroup}},
+		getGeneric{g.client.(*pango.Firewall).Network.BgpPeer, []string{virtualRouter, bgpPeerGroup}},
 		virtualRouter+":"+bgpPeerGroup+":", true, "panos_bgp_peer", false, "",
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createBGPRedistResources(virtualRouter string) []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.BgpRedistRule, []string{virtualRouter}},
+		getGeneric{g.client.(*pango.Firewall).Network.BgpRedistRule, []string{virtualRouter}},
 		virtualRouter+":", true, "panos_bgp_redist_rule", false, "",
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createEthernetInterfaceResources() (resources []terraformutils.Resource) {
-	l, err := g.client.Network.EthernetInterface.GetList()
+	l, err := g.client.(*pango.Firewall).Network.EthernetInterface.GetList()
 	if err != nil {
 		return []terraformutils.Resource{}
 	}
 
 	for _, ethernetInterface := range l {
-		rv, err := g.client.IsImported(util.InterfaceImport, "", "", g.vsys, ethernetInterface)
+		rv, err := g.client.(*pango.Firewall).IsImported(util.InterfaceImport, "", "", g.vsys, ethernetInterface)
 		if err != nil || !rv {
 			continue
 		}
@@ -290,7 +322,7 @@ func (g *FirewallNetworkingGenerator) createEthernetInterfaceResources() (resour
 			[]string{},
 		))
 
-		e, err := g.client.Network.EthernetInterface.Get(ethernetInterface)
+		e, err := g.client.(*pango.Firewall).Network.EthernetInterface.Get(ethernetInterface)
 		if err != nil {
 			continue
 		}
@@ -309,13 +341,13 @@ func (g *FirewallNetworkingGenerator) createEthernetInterfaceResources() (resour
 
 func (g *FirewallNetworkingGenerator) createGRETunnelResources() []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.GreTunnel, []string{}},
+		getGeneric{g.client.(*pango.Firewall).Network.GreTunnel, []string{}},
 		"", false, "panos_gre_tunnel", false, "",
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createIKECryptoProfileResources() (resources []terraformutils.Resource) {
-	l, err := g.client.Network.IkeCryptoProfile.GetList()
+	l, err := g.client.(*pango.Firewall).Network.IkeCryptoProfile.GetList()
 	if err != nil {
 		return []terraformutils.Resource{}
 	}
@@ -338,7 +370,7 @@ func (g *FirewallNetworkingGenerator) createIKECryptoProfileResources() (resourc
 }
 
 func (g *FirewallNetworkingGenerator) createIKEGatewayResources() (resources []terraformutils.Resource) {
-	l, err := g.client.Network.IkeGateway.GetList()
+	l, err := g.client.(*pango.Firewall).Network.IkeGateway.GetList()
 	if err != nil {
 		return []terraformutils.Resource{}
 	}
@@ -361,7 +393,7 @@ func (g *FirewallNetworkingGenerator) createIKEGatewayResources() (resources []t
 }
 
 func (g *FirewallNetworkingGenerator) createIPSECCryptoProfileResources() (resources []terraformutils.Resource) {
-	l, err := g.client.Network.IpsecCryptoProfile.GetList()
+	l, err := g.client.(*pango.Firewall).Network.IpsecCryptoProfile.GetList()
 	if err != nil {
 		return []terraformutils.Resource{}
 	}
@@ -385,13 +417,13 @@ func (g *FirewallNetworkingGenerator) createIPSECCryptoProfileResources() (resou
 
 func (g *FirewallNetworkingGenerator) createIPSECTunnelProxyIDIPv4Resources(ipsecTunnel string) []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.IpsecTunnelProxyId, []string{ipsecTunnel}},
+		getGeneric{g.client.(*pango.Firewall).Network.IpsecTunnelProxyId, []string{ipsecTunnel}},
 		ipsecTunnel+":", false, "panos_ipsec_tunnel_proxy_id_ipv4", false, "",
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createIPSECTunnelResources() (resources []terraformutils.Resource) {
-	l, err := g.client.Network.IpsecTunnel.GetList()
+	l, err := g.client.(*pango.Firewall).Network.IpsecTunnel.GetList()
 	if err != nil {
 		return []terraformutils.Resource{}
 	}
@@ -413,27 +445,27 @@ func (g *FirewallNetworkingGenerator) createIPSECTunnelResources() (resources []
 
 func (g *FirewallNetworkingGenerator) createLayer2SubInterfaceResources(interfaceType, parentInterface, parentMode string) []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.Layer2Subinterface, []string{interfaceType, parentInterface, parentMode}},
+		getGeneric{g.client.(*pango.Firewall).Network.Layer2Subinterface, []string{interfaceType, parentInterface, parentMode}},
 		interfaceType+":"+parentInterface+":"+parentMode+":"+g.vsys+":", false, "panos_layer2_subinterface", true, util.InterfaceImport,
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createLayer3SubInterfaceResources(interfaceType, parentInterface string) []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.Layer3Subinterface, []string{interfaceType, parentInterface}},
+		getGeneric{g.client.(*pango.Firewall).Network.Layer3Subinterface, []string{interfaceType, parentInterface}},
 		interfaceType+":"+parentInterface+":"+g.vsys+":", false, "panos_layer3_subinterface", true, util.InterfaceImport,
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createLoopbackInterfaceResources() []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.LoopbackInterface, []string{}},
+		getGeneric{g.client.(*pango.Firewall).Network.LoopbackInterface, []string{}},
 		g.vsys+":", false, "panos_loopback_interface", true, util.InterfaceImport,
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createManagementProfileResources() (resources []terraformutils.Resource) {
-	l, err := g.client.Network.ManagementProfile.GetList()
+	l, err := g.client.(*pango.Firewall).Network.ManagementProfile.GetList()
 	if err != nil {
 		return []terraformutils.Resource{}
 	}
@@ -457,41 +489,41 @@ func (g *FirewallNetworkingGenerator) createManagementProfileResources() (resour
 
 func (g *FirewallNetworkingGenerator) createMonitorProfileResources() []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.MonitorProfile, []string{}},
+		getGeneric{g.client.(*pango.Firewall).Network.MonitorProfile, []string{}},
 		"", false, "panos_monitor_profile", false, "",
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createRedistributionProfileResources(virtualRouter string) []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.RedistributionProfile, []string{virtualRouter}},
+		getGeneric{g.client.(*pango.Firewall).Network.RedistributionProfile, []string{virtualRouter}},
 		virtualRouter+":", true, "panos_redistribution_profile_ipv4", false, "",
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createStaticRouteIpv4Resources(virtualRouter string) []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.StaticRoute, []string{virtualRouter}},
+		getGeneric{g.client.(*pango.Firewall).Network.StaticRoute, []string{virtualRouter}},
 		virtualRouter+":", true, "panos_static_route_ipv4", false, "",
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createTunnelInterfaceResources() []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.TunnelInterface, []string{}},
+		getGeneric{g.client.(*pango.Firewall).Network.TunnelInterface, []string{}},
 		g.vsys+":", false, "panos_tunnel_interface", true, util.InterfaceImport,
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createVirtualRouterResources() (resources []terraformutils.Resource) {
-	l, err := g.client.Network.VirtualRouter.GetList()
+	l, err := g.client.(*pango.Firewall).Network.VirtualRouter.GetList()
 	if err != nil {
 		return []terraformutils.Resource{}
 	}
 
 	for _, virtualRouter := range l {
 		// TODO: doesn't work!!?
-		// rv, err := g.client.IsImported(util.VirtualRouterImport, "", "", g.vsys, virtualRouter)
+		// rv, err := g.client.(*pango.Firewall).IsImported(util.VirtualRouterImport, "", "", g.vsys, virtualRouter)
 		// if err != nil || !rv {
 		// 	continue
 		// }
@@ -524,21 +556,21 @@ func (g *FirewallNetworkingGenerator) createVirtualRouterResources() (resources
 func (g *FirewallNetworkingGenerator) createVlanResources() []terraformutils.Resource {
 	// TODO: should activate check with util.VlanImport, but doesn't work?
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.Vlan, []string{}},
+		getGeneric{g.client.(*pango.Firewall).Network.Vlan, []string{}},
 		g.vsys+":", false, "panos_vlan", false, "",
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createVlanInterfaceResources() []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.VlanInterface, []string{}},
+		getGeneric{g.client.(*pango.Firewall).Network.VlanInterface, []string{}},
 		g.vsys+":", false, "panos_vlan_interface", true, util.InterfaceImport,
 	)
 }
 
 func (g *FirewallNetworkingGenerator) createZoneResources() []terraformutils.Resource {
 	return g.createResourcesFromList(
-		getGeneric{g.client.Network.Zone, []string{g.vsys}},
+		getGeneric{g.client.(*pango.Firewall).Network.Zone, []string{g.vsys}},
 		g.vsys+":", false, "panos_zone", false, "",
 	)
 }
@@ -645,7 +677,6 @@ func (g *FirewallNetworkingGenerator) PostConvertHook() error {
 
 		if r.InstanceInfo.Type == "panos_bgp_peer" {
 			r.Item["virtual_router"] = "${panos_bgp." + normalizeResourceName(r.Item["virtual_router"].(string)) + ".virtual_router}"
-			r.Item["bgp_peer_group"] = "${panos_bgp_peer_group." + normalizeResourceName(r.Item["panos_bgp_peer_group"].(string)) + ".name}"
 			r.Item["peer_as"] = "${panos_bgp." + normalizeResourceName(r.Item["virtual_router"].(string)) + ".as_number}"
 		}
 
@@ -669,7 +700,9 @@ func (g *FirewallNetworkingGenerator) PostConvertHook() error {
 		if r.InstanceInfo.Type == "panos_ipsec_tunnel" {
 			r.Item["tunnel_interface"] = mapInterfaceNames[r.Item["tunnel_interface"].(string)]
 			r.Item["ak_ike_gateway"] = mapIKEGatewayNames[r.Item["ak_ike_gateway"].(string)]
-			r.Item["ak_ipsec_crypto_profile"] = mapIPSECCryptoProfileNames[r.Item["ak_ipsec_crypto_profile"].(string)]
+			if _, ok := r.Item["ak_ipsec_crypto_profile"]; ok {
+				r.Item["ak_ipsec_crypto_profile"] = mapIPSECCryptoProfileNames[r.Item["ak_ipsec_crypto_profile"].(string)]
+			}
 		}
 
 		if r.InstanceInfo.Type == "panos_ipsec_tunnel_proxy_id_ipv4" {
